"""Classes and functions used to test the workbench.

These classes are called by the unit test launcher
that is defined in `Init.py` and `InitGui.py`.

The unit tests are based on the standard `unittest` module.
"""
