"""Classes that define the task panels of GUI commands.

These classes load `.ui` files that will be used in the task panel
of the graphical commands.
The classes define the behavior and callbacks of the different widgets
included in the `.ui` file.
"""
