# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2016 - Victor Titov (DeepSOIC) <vv.titov@gmail.com>     *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "ImageTools package"  # from "Macro Image Scaling"
__author__  = "JAndersM"
__url__     = "https://www.freecadweb.org/wiki/Macro_Image_Scaling"
__version__ = "00.01"
__date__    = "19/01/2016" 
 
 
__tutorial__ = "https://youtu.be/2iFE40uHrA8" 
__forum__ = "https://forum.freecadweb.org/viewtopic.php?f=3&t=14265"


## @package ImageTools
#  \ingroup Image
#  \brief ImageTools Package for Image workbench
