body {
    background: BACKGROUND;
    color: BGTCOLOR;
    font-family: FONTFAMILY;
    font-size: FONTSIZE;
}
a, a:link, a:visited {
    text-decoration: none;
    color: LINKCOLOR;
    font-weight: bold;
}
h1 {
    width: 100%;
    text-align: center;
}
h2 {
    text-align: center;
    clear: both;
}
hr {
    height: 20px;
    border: none;
}
.hidden {
    display: none;
}
.version {
    font-size: 0.8em;
    float: right;
    padding-right: 10px;
}
.settingsicon {
    width: 16px;
    height: 16px;
    margin-left: 20px;
    vertical-align:middle;
}
.container {
    margin: 2%
}
ul.tabs {
    margin: 9px 0;
}
ul.tabs li {
    list-style: none;
    display: inline;
}
ul.tabs li a {
    color: TEXTCOLOR;
    background-color: BOXCOLOR;
    padding: 8px 14px 8px 14px;
    text-decoration: none;
    font-size: 1em;
    font-weight: bold;
    text-transform: uppercase;
    border:1px solid BOXCOLOR;
    border-radius: 3px 3px 0px 0px;
    box-shadow: #445 0 -5px 5px;
    outline: 0;
}
ul.tabs li a:hover {
    background-color: TEXTCOLOR;
    color: BASECOLOR;
    border-color: TEXTCOLOR;
}
ul.tabs li a.active, ul.tabs li a.active:hover {
    color: TEXTCOLOR;
    background-color: BASECOLOR;
    border-bottom: 1px solid BASECOLOR;
    border-color: BASECOLOR;
}
.panel {
    background: BASECOLOR;
    color: TEXTCOLOR;
    padding: 15px;
    clear: both;
    margin-bottom: 10px;
    border-radius: 0px 3px 3px 3px;
    overflow: auto;
    box-shadow: SHADOW 0 2px 5px;
}
ul.icons {
    clear: both;
}
.icon {
    list-style: none;
    display: inline;
    float: left;
    padding: 5px 0px 15px 10px;
    width: 138px;
    height: 200px;
    background: BOXCOLOR;
    border-radius: 3px;
    margin: 10px;
    box-shadow: SHADOW 0 0px 6px;
    word-wrap: break-word;
}
.icon h4 {
    margin: 3px 0;
    max-width: 90%;
}
.icon img {
    width: 128px;
    height: 128px;
}
.icon p {
    margin: 0;
}
.caption {
    clear: both;
}
a .caption, a .caption:link, a .caption:visited {
    color: TEXTCOLOR;
    text-decoration-color: TEXTCOLOR;
}

.caption p {
    font-weight: normal;
}

.docbox a:hover {
    text-decoration: underline;
}
.docbox {
    float: left;
    padding: 10px;
    max-width: 360px;
    background: BOXCOLOR;
    border-radius: 3px;
    margin: 10px;
    box-shadow: SHADOW 0 0px 6px;
}
.docbox h2 {
    text-align: center;
}
.docbox img {
    max-width: 24px;
    max-height: 24px;
}
.forum img {
    max-width: 100% !important;
    max-height: 100% !important;
}
.allwidth {
    max-width: 98%;
}
h3 a {
    height: 32px;
    vertical-align: middle;
    padding-bottom: 18px;
    padding-left: 5px;
}
#commits ul {
    margin: 30px 0;
}
#commits ul li, .forum ul li {
    margin-bottom: 10px;
}
.footnote {
    text-align: center;
    display: block; /* footnote tips display */
    clear: both;
    padding-top: 10px;
}
ul.workbenches, ul.addonslist {
    overflow: hidden;
    padding: 10px;
}
ul.workbenches li, ul.addonslist li {
    list-style: none;
    float: left;
    display: inline;
    width: 50%;
}
.forum {
    display: none; /* forum display */ /* don't change this line */
}
.forum pre {
    white-space: pre-wrap;
}
.thumbnails {
    float: left;
    width: 100%; /* thumbs display */ /* don't change this line */
}
.notes {
    display: none; /* notes display */ /* don't change this line */
    float: right;
    width: 30%;
}
#notepad {
    width: 100%;
    height: 100%;
    color: TEXTCOLOR;
    background: BOXCOLOR;
    border: none;
    padding: 5px;
    border-radius: 3px;
}
