# -*- coding: utf-8 -*-

# Resource object code
#
# Created: Sun May 3 20:14:20 2020
#      by: The Resource Compiler for PySide2 (Qt v5.12.5)
#
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore

qt_resource_data = b"\
\x00\x00\x17]\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0d\x0a<!-- Crea\
ted with Inkscap\
e (http://www.in\
kscape.org/) -->\
\x0d\x0a\x0d\x0a<svg\x0d\x0a   xml\
ns:dc=\x22http://pu\
rl.org/dc/elemen\
ts/1.1/\x22\x0d\x0a   xml\
ns:cc=\x22http://cr\
eativecommons.or\
g/ns#\x22\x0d\x0a   xmlns\
:rdf=\x22http://www\
.w3.org/1999/02/\
22-rdf-syntax-ns\
#\x22\x0d\x0a   xmlns:svg\
=\x22http://www.w3.\
org/2000/svg\x22\x0d\x0a \
  xmlns=\x22http://\
www.w3.org/2000/\
svg\x22\x0d\x0a   xmlns:x\
link=\x22http://www\
.w3.org/1999/xli\
nk\x22\x0d\x0a   xmlns:so\
dipodi=\x22http://s\
odipodi.sourcefo\
rge.net/DTD/sodi\
podi-0.dtd\x22\x0d\x0a   \
xmlns:inkscape=\x22\
http://www.inksc\
ape.org/namespac\
es/inkscape\x22\x0d\x0a  \
 width=\x2264\x22\x0d\x0a   \
height=\x2264\x22\x0d\x0a   \
id=\x22svg2\x22\x0d\x0a   ve\
rsion=\x221.1\x22\x0d\x0a   \
inkscape:version\
=\x220.48.5 r10040\x22\
\x0d\x0a   sodipodi:do\
cname=\x22preview-v\
ector.svg\x22>\x0d\x0a  <\
defs\x0d\x0a     id=\x22d\
efs4\x22>\x0d\x0a    <ink\
scape:perspectiv\
e\x0d\x0a       sodipo\
di:type=\x22inkscap\
e:persp3d\x22\x0d\x0a    \
   inkscape:vp_x\
=\x22-206.25075 : 1\
19.68854 : 1\x22\x0d\x0a \
      inkscape:v\
p_y=\x220 : 1000 : \
0\x22\x0d\x0a       inksc\
ape:vp_z=\x22183.50\
001 : 140.32143 \
: 1\x22\x0d\x0a       ink\
scape:persp3d-or\
igin=\x2248 : 32 : \
1\x22\x0d\x0a       id=\x22p\
erspective3792\x22 \
/>\x0d\x0a    <linearG\
radient\x0d\x0a       \
id=\x22linearGradie\
nt3755\x22>\x0d\x0a      \
<stop\x0d\x0a         \
style=\x22stop-colo\
r:#424242;stop-o\
pacity:1;\x22\x0d\x0a    \
     offset=\x220\x22\x0d\
\x0a         id=\x22st\
op3757\x22 />\x0d\x0a    \
  <stop\x0d\x0a       \
  style=\x22stop-co\
lor:#9f9f9f;stop\
-opacity:1;\x22\x0d\x0a  \
       offset=\x221\
\x22\x0d\x0a         id=\x22\
stop3759\x22 />\x0d\x0a  \
  </linearGradie\
nt>\x0d\x0a    <linear\
Gradient\x0d\x0a      \
 inkscape:collec\
t=\x22always\x22\x0d\x0a    \
   xlink:href=\x22#\
linearGradient37\
55\x22\x0d\x0a       id=\x22\
linearGradient37\
61\x22\x0d\x0a       x1=\x22\
75.724594\x22\x0d\x0a    \
   y1=\x2285.023956\
\x22\x0d\x0a       x2=\x2241\
.895164\x22\x0d\x0a      \
 y2=\x2213.334851\x22\x0d\
\x0a       gradient\
Units=\x22userSpace\
OnUse\x22 />\x0d\x0a  </d\
efs>\x0d\x0a  <sodipod\
i:namedview\x0d\x0a   \
  id=\x22base\x22\x0d\x0a   \
  pagecolor=\x22#ff\
ffff\x22\x0d\x0a     bord\
ercolor=\x22#666666\
\x22\x0d\x0a     borderop\
acity=\x221.0\x22\x0d\x0a   \
  inkscape:pageo\
pacity=\x220.0\x22\x0d\x0a  \
   inkscape:page\
shadow=\x222\x22\x0d\x0a    \
 inkscape:zoom=\x22\
8.7571168\x22\x0d\x0a    \
 inkscape:cx=\x2239\
.859065\x22\x0d\x0a     i\
nkscape:cy=\x2230.9\
25664\x22\x0d\x0a     ink\
scape:document-u\
nits=\x22px\x22\x0d\x0a     \
inkscape:current\
-layer=\x22g3794\x22\x0d\x0a\
     showgrid=\x22t\
rue\x22\x0d\x0a     showb\
order=\x22true\x22\x0d\x0a  \
   inkscape:wind\
ow-width=\x22800\x22\x0d\x0a\
     inkscape:wi\
ndow-height=\x22836\
\x22\x0d\x0a     inkscape\
:window-x=\x220\x22\x0d\x0a \
    inkscape:win\
dow-y=\x2227\x22\x0d\x0a    \
 inkscape:window\
-maximized=\x220\x22\x0d\x0a\
     inkscape:sn\
ap-bbox=\x22false\x22\x0d\
\x0a     inkscape:s\
nap-nodes=\x22true\x22\
\x0d\x0a     inkscape:\
snap-global=\x22tru\
e\x22>\x0d\x0a    <inksca\
pe:grid\x0d\x0a       \
type=\x22xygrid\x22\x0d\x0a \
      id=\x22grid30\
02\x22\x0d\x0a       emps\
pacing=\x222\x22\x0d\x0a    \
   visible=\x22true\
\x22\x0d\x0a       enable\
d=\x22true\x22\x0d\x0a      \
 snapvisiblegrid\
linesonly=\x22true\x22\
 />\x0d\x0a  </sodipod\
i:namedview>\x0d\x0a  \
<metadata\x0d\x0a     \
id=\x22metadata7\x22>\x0d\
\x0a    <rdf:RDF>\x0d\x0a\
      <cc:Work\x0d\x0a\
         rdf:abo\
ut=\x22\x22>\x0d\x0a        \
<dc:format>image\
/svg+xml</dc:for\
mat>\x0d\x0a        <d\
c:type\x0d\x0a        \
   rdf:resource=\
\x22http://purl.org\
/dc/dcmitype/Sti\
llImage\x22 />\x0d\x0a   \
     <dc:title /\
>\x0d\x0a        <dc:c\
reator>\x0d\x0a       \
   <cc:Agent>\x0d\x0a \
           <dc:t\
itle>$committer<\
/dc:title>\x0d\x0a    \
      </cc:Agent\
>\x0d\x0a        </dc:\
creator>\x0d\x0a      \
  <dc:title>prev\
iew-vector</dc:t\
itle>\x0d\x0a        <\
dc:date>2013-11-\
19</dc:date>\x0d\x0a  \
      <dc:relati\
on>http://www.fr\
eecadweb.org/wik\
i/index.php?titl\
e=Artwork</dc:re\
lation>\x0d\x0a       \
 <dc:publisher>\x0d\
\x0a          <cc:A\
gent>\x0d\x0a         \
   <dc:title>Fre\
eCAD</dc:title>\x0d\
\x0a          </cc:\
Agent>\x0d\x0a        \
</dc:publisher>\x0d\
\x0a        <dc:ide\
ntifier>FreeCAD/\
src/Mod/Material\
/Resources/icons\
/preview-vector.\
svg</dc:identifi\
er>\x0d\x0a        <dc\
:rights>\x0d\x0a      \
    <cc:Agent>\x0d\x0a\
            <dc:\
title>FreeCAD LG\
PL2+</dc:title>\x0d\
\x0a          </cc:\
Agent>\x0d\x0a        \
</dc:rights>\x0d\x0a  \
      <cc:licens\
e>https://www.gn\
u.org/copyleft/l\
esser.html</cc:l\
icense>\x0d\x0a       \
 <dc:contributor\
>\x0d\x0a          <cc\
:Agent>\x0d\x0a       \
     <dc:title>[\
agryson] Alexand\
er Gryson</dc:ti\
tle>\x0d\x0a          \
</cc:Agent>\x0d\x0a   \
     </dc:contri\
butor>\x0d\x0a      </\
cc:Work>\x0d\x0a    </\
rdf:RDF>\x0d\x0a  </me\
tadata>\x0d\x0a  <g\x0d\x0a \
    inkscape:lab\
el=\x22Layer 1\x22\x0d\x0a  \
   inkscape:grou\
pmode=\x22layer\x22\x0d\x0a \
    id=\x22layer1\x22\x0d\
\x0a     transform=\
\x22translate(0,-98\
8.36218)\x22>\x0d\x0a    \
<g\x0d\x0a       id=\x22g\
3794\x22\x0d\x0a       st\
yle=\x22color:#0000\
00;fill:#ffffff;\
fill-opacity:1;f\
ill-rule:evenodd\
;stroke:none;str\
oke-width:3.2939\
9991;marker:none\
;visibility:visi\
ble;display:inli\
ne;overflow:visi\
ble;enable-backg\
round:accumulate\
\x22>\x0d\x0a      <path\x0d\
\x0a         style=\
\x22fill:#d3d7cf;fi\
ll-opacity:1;fil\
l-rule:evenodd;s\
troke:#000000;st\
roke-width:2;str\
oke-linecap:roun\
d;stroke-linejoi\
n:round;stroke-m\
iterlimit:4;stro\
ke-opacity:1;str\
oke-dasharray:no\
ne\x22\x0d\x0a         d=\
\x22m 29.814326,991\
.36218 -8.814326\
,4 -1,7.00002 26\
,1 14.994091,-8.\
6577 z\x22\x0d\x0a       \
  id=\x22path3816\x22\x0d\
\x0a         inksca\
pe:connector-cur\
vature=\x220\x22\x0d\x0a    \
     sodipodi:no\
detypes=\x22cccccc\x22\
 />\x0d\x0a      <path\
\x0d\x0a         d=\x22M \
38.484871,1007.8\
171 40,1046.3622\
 l 21,-22.4044 0\
,-29.25647 z\x22\x0d\x0a \
        style=\x22f\
ill:#b7b7b7;fill\
-opacity:1;fill-\
rule:evenodd;str\
oke:#000000;stro\
ke-width:2;strok\
e-linecap:round;\
stroke-linejoin:\
round;stroke-mit\
erlimit:4;stroke\
-opacity:1;strok\
e-dasharray:none\
\x22\x0d\x0a         id=\x22\
path3818\x22\x0d\x0a     \
    inkscape:con\
nector-curvature\
=\x220\x22\x0d\x0a         s\
odipodi:nodetype\
s=\x22ccccc\x22 />\x0d\x0a  \
    <path\x0d\x0a     \
    style=\x22fill:\
#babdb6;stroke:#\
d3d7cf;stroke-wi\
dth:2;stroke-lin\
ecap:butt;stroke\
-linejoin:miter;\
stroke-opacity:1\
\x22\x0d\x0a         d=\x22M\
 59.035745,9.662\
7807 59.108006,3\
4.818961 40.9153\
1,54.264958 42.6\
98526,19.372964 \
z\x22\x0d\x0a         id=\
\x22path3022\x22\x0d\x0a    \
     inkscape:co\
nnector-curvatur\
e=\x220\x22\x0d\x0a         \
transform=\x22trans\
late(0,988.36218\
)\x22\x0d\x0a         sod\
ipodi:nodetypes=\
\x22ccccc\x22 />\x0d\x0a    \
  <path\x0d\x0a       \
  d=\x22m 21,995.36\
218 31,5.00002 -\
12,46 -30,-8 z\x22\x0d\
\x0a         style=\
\x22fill:#ffffff;fi\
ll-opacity:1;fil\
l-rule:evenodd;s\
troke:#000000;st\
roke-width:4;str\
oke-linejoin:rou\
nd;stroke-miterl\
imit:4;stroke-op\
acity:1;stroke-d\
asharray:none\x22\x0d\x0a\
         id=\x22pat\
h3820\x22\x0d\x0a        \
 inkscape:connec\
tor-curvature=\x220\
\x22\x0d\x0a         sodi\
podi:nodetypes=\x22\
ccccc\x22 />\x0d\x0a     \
 <path\x0d\x0a        \
 d=\x22m 21,995.362\
18 31,5.00002 -1\
2,46 -30,-8 z\x22\x0d\x0a\
         style=\x22\
color:#000000;fi\
ll:none;stroke:#\
888a85;stroke-wi\
dth:2;stroke-lin\
ejoin:round;stro\
ke-miterlimit:4;\
stroke-opacity:1\
;stroke-dasharra\
y:none;marker:no\
ne;visibility:vi\
sible;display:in\
line;overflow:vi\
sible;enable-bac\
kground:accumula\
te\x22\x0d\x0a         id\
=\x22path3820-3\x22\x0d\x0a \
        inkscape\
:connector-curva\
ture=\x220\x22\x0d\x0a      \
   sodipodi:node\
types=\x22ccccc\x22 />\
\x0d\x0a    </g>\x0d\x0a  </\
g>\x0d\x0a</svg>\x0d\x0a\
\x00\x00\x18\xe0\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0d\x0a<!-- Crea\
ted with Inkscap\
e (http://www.in\
kscape.org/) -->\
\x0d\x0a<svg xmlns:dc=\
\x22http://purl.org\
/dc/elements/1.1\
/\x22 xmlns:cc=\x22htt\
p://creativecomm\
ons.org/ns#\x22 xml\
ns:rdf=\x22http://w\
ww.w3.org/1999/0\
2/22-rdf-syntax-\
ns#\x22 xmlns:svg=\x22\
http://www.w3.or\
g/2000/svg\x22 xmln\
s=\x22http://www.w3\
.org/2000/svg\x22 x\
mlns:xlink=\x22http\
://www.w3.org/19\
99/xlink\x22 xmlns:\
sodipodi=\x22http:/\
/sodipodi.source\
forge.net/DTD/so\
dipodi-0.dtd\x22 xm\
lns:inkscape=\x22ht\
tp://www.inkscap\
e.org/namespaces\
/inkscape\x22 width\
=\x2264\x22 height=\x2264\
\x22 id=\x22svg2\x22 vers\
ion=\x221.1\x22 inksca\
pe:version=\x220.48\
.5 r10040\x22 sodip\
odi:docname=\x22pre\
view-rendered.sv\
g\x22>\x0d\x0a  <defs id=\
\x22defs4\x22>\x0d\x0a    <l\
inearGradient id\
=\x22linearGradient\
3755\x22>\x0d\x0a      <s\
top style=\x22stop-\
color:#424242;st\
op-opacity:1;\x22 o\
ffset=\x220\x22 id=\x22st\
op3757\x22/>\x0d\x0a     \
 <stop style=\x22st\
op-color:#9f9f9f\
;stop-opacity:1;\
\x22 offset=\x221\x22 id=\
\x22stop3759\x22/>\x0d\x0a  \
  </linearGradie\
nt>\x0d\x0a    <filter\
 color-interpola\
tion-filters=\x22sR\
GB\x22 inkscape:col\
lect=\x22always\x22 id\
=\x22filter3871\x22 x=\
\x22-0.45363569\x22 wi\
dth=\x221.9072714\x22 \
y=\x22-0.45363569\x22 \
height=\x221.907271\
4\x22>\x0d\x0a      <feGa\
ussianBlur inksc\
ape:collect=\x22alw\
ays\x22 stdDeviatio\
n=\x229.8975056\x22 id\
=\x22feGaussianBlur\
3873\x22/>\x0d\x0a    </f\
ilter>\x0d\x0a    <lin\
earGradient inks\
cape:collect=\x22al\
ways\x22 xlink:href\
=\x22#linearGradien\
t3600\x22 id=\x22linea\
rGradient3606\x22 x\
1=\x2251.037281\x22 y1\
=\x2247.692612\x22 x2=\
\x2214.872466\x22 y2=\x22\
10.397644\x22 gradi\
entUnits=\x22userSp\
aceOnUse\x22/>\x0d\x0a   \
 <linearGradient\
 id=\x22linearGradi\
ent3600\x22>\x0d\x0a     \
 <stop style=\x22st\
op-color:#555753\
;stop-opacity:1\x22\
 offset=\x220\x22 id=\x22\
stop3602\x22/>\x0d\x0a   \
   <stop style=\x22\
stop-color:#d3d7\
cf;stop-opacity:\
1\x22 offset=\x221\x22 id\
=\x22stop3604\x22/>\x0d\x0a \
   </linearGradi\
ent>\x0d\x0a    <filte\
r inkscape:colle\
ct=\x22always\x22 id=\x22\
filter3805\x22 x=\x22-\
0.29648924\x22 widt\
h=\x221.5929785\x22 y=\
\x22-0.30309633\x22 he\
ight=\x221.6061927\x22\
>\x0d\x0a      <feGaus\
sianBlur inkscap\
e:collect=\x22alway\
s\x22 stdDeviation=\
\x222.2490136\x22 id=\x22\
feGaussianBlur38\
07\x22/>\x0d\x0a    </fil\
ter>\x0d\x0a  </defs>\x0d\
\x0a  <sodipodi:nam\
edview id=\x22base\x22\
 pagecolor=\x22#fff\
fff\x22 bordercolor\
=\x22#666666\x22 borde\
ropacity=\x221.0\x22 i\
nkscape:pageopac\
ity=\x220.0\x22 inksca\
pe:pageshadow=\x222\
\x22 inkscape:zoom=\
\x224.8193618\x22 inks\
cape:cx=\x22-4.0027\
486\x22 inkscape:cy\
=\x2228.204234\x22 ink\
scape:document-u\
nits=\x22px\x22 inksca\
pe:current-layer\
=\x22layer1\x22 showgr\
id=\x22true\x22 showbo\
rder=\x22true\x22 inks\
cape:window-widt\
h=\x22800\x22 inkscape\
:window-height=\x22\
837\x22 inkscape:wi\
ndow-x=\x220\x22 inksc\
ape:window-y=\x2227\
\x22 inkscape:windo\
w-maximized=\x220\x22>\
\x0d\x0a    <inkscape:\
grid type=\x22xygri\
d\x22 id=\x22grid2988\x22\
 empspacing=\x222\x22 \
visible=\x22true\x22 e\
nabled=\x22true\x22 sn\
apvisiblegridlin\
esonly=\x22true\x22/>\x0d\
\x0a  </sodipodi:na\
medview>\x0d\x0a  <met\
adata id=\x22metada\
ta7\x22>\x0d\x0a    <rdf:\
RDF>\x0d\x0a      <cc:\
Work rdf:about=\x22\
\x22>\x0d\x0a        <dc:\
format>image/svg\
+xml</dc:format>\
\x0d\x0a        <dc:ty\
pe rdf:resource=\
\x22http://purl.org\
/dc/dcmitype/Sti\
llImage\x22/>\x0d\x0a    \
    <dc:title/>\x0d\
\x0a        <dc:cre\
ator>\x0d\x0a         \
 <cc:Agent>\x0d\x0a   \
         <dc:tit\
le>$committer</d\
c:title>\x0d\x0a      \
    </cc:Agent>\x0d\
\x0a        </dc:cr\
eator>\x0d\x0a        \
<dc:title>previe\
w-rendered</dc:t\
itle>\x0d\x0a        <\
dc:date>2013-11-\
19</dc:date>\x0d\x0a  \
      <dc:relati\
on>http://www.fr\
eecadweb.org/wik\
i/index.php?titl\
e=Artwork</dc:re\
lation>\x0d\x0a       \
 <dc:publisher>\x0d\
\x0a          <cc:A\
gent>\x0d\x0a         \
   <dc:title>Fre\
eCAD</dc:title>\x0d\
\x0a          </cc:\
Agent>\x0d\x0a        \
</dc:publisher>\x0d\
\x0a        <dc:ide\
ntifier>FreeCAD/\
src/Mod/Material\
/Resources/icons\
/preview-rendere\
d.svg</dc:identi\
fier>\x0d\x0a        <\
dc:rights>\x0d\x0a    \
      <cc:Agent>\
\x0d\x0a            <d\
c:title>FreeCAD \
LGPL2+</dc:title\
>\x0d\x0a          </c\
c:Agent>\x0d\x0a      \
  </dc:rights>\x0d\x0a\
        <cc:lice\
nse>https://www.\
gnu.org/copyleft\
/lesser.html</cc\
:license>\x0d\x0a     \
   <dc:contribut\
or>\x0d\x0a          <\
cc:Agent>\x0d\x0a     \
       <dc:title\
>[agryson] Alexa\
nder Gryson</dc:\
title>\x0d\x0a        \
  </cc:Agent>\x0d\x0a \
       </dc:cont\
ributor>\x0d\x0a      \
</cc:Work>\x0d\x0a    \
</rdf:RDF>\x0d\x0a  </\
metadata>\x0d\x0a  <g \
inkscape:label=\x22\
Layer 1\x22 inkscap\
e:groupmode=\x22lay\
er\x22 id=\x22layer1\x22 \
transform=\x22trans\
late(0,-988.3621\
8)\x22>\x0d\x0a    <path \
sodipodi:type=\x22a\
rc\x22 style=\x22color\
:#000000;fill:#0\
00000;fill-opaci\
ty:0.65625;fill-\
rule:evenodd;str\
oke:none;stroke-\
width:5;marker:n\
one;visibility:v\
isible;display:i\
nline;overflow:v\
isible;filter:ur\
l(#filter3871);e\
nable-background\
:accumulate\x22 id=\
\x22path2826-8\x22 sod\
ipodi:cx=\x2231.636\
364\x22 sodipodi:cy\
=\x2230.363636\x22 sod\
ipodi:rx=\x2226.181\
818\x22 sodipodi:ry\
=\x2226.181818\x22 d=\x22\
m 57.818182,30.3\
63636 a 26.18181\
8,26.181818 0 1 \
1 -52.363636,0 2\
6.181818,26.1818\
18 0 1 1 52.3636\
36,0 z\x22 transfor\
m=\x22matrix(0.4806\
1681,0,0,0.12015\
42,24.795032,104\
2.7139)\x22/>\x0d\x0a    \
<path sodipodi:t\
ype=\x22arc\x22 style=\
\x22color:#000000;f\
ill:url(#linearG\
radient3606);fil\
l-opacity:1;fill\
-rule:evenodd;st\
roke:#0b1521;str\
oke-width:1.8056\
4272;marker:none\
;visibility:visi\
ble;display:inli\
ne;overflow:visi\
ble;enable-backg\
round:accumulate\
\x22 id=\x22path2826\x22 \
sodipodi:cx=\x2231.\
636364\x22 sodipodi\
:cy=\x2230.363636\x22 \
sodipodi:rx=\x2226.\
181818\x22 sodipodi\
:ry=\x2226.181818\x22 \
d=\x22m 57.818182,3\
0.363636 a 26.18\
1818,26.181818 0\
 1 1 -52.363636,\
0 26.181818,26.1\
81818 0 1 1 52.3\
63636,0 z\x22 trans\
form=\x22matrix(1.1\
076389,0,0,1.107\
6389,-3.0416672,\
986.73024)\x22/>\x0d\x0a \
   <path sodipod\
i:type=\x22arc\x22 sty\
le=\x22color:#00000\
0;fill:none;stro\
ke:#d3d7cf;strok\
e-width:1.939394\
12000000000;mark\
er:none;visibili\
ty:visible;displ\
ay:inline;overfl\
ow:visible;enabl\
e-background:acc\
umulate\x22 id=\x22pat\
h2826-4\x22 sodipod\
i:cx=\x2231.636364\x22\
 sodipodi:cy=\x2230\
.363636\x22 sodipod\
i:rx=\x2226.181818\x22\
 sodipodi:ry=\x2226\
.181818\x22 d=\x22m 57\
.818182,30.36363\
6 a 26.181818,26\
.181818 0 1 1 -5\
2.363636,0 26.18\
1818,26.181818 0\
 1 1 52.363636,0\
 z\x22 transform=\x22m\
atrix(1.03125,0,\
0,1.03125,-0.625\
00046,989.04968)\
\x22/>\x0d\x0a    <path s\
odipodi:type=\x22ar\
c\x22 style=\x22opacit\
y:0.844827569999\
99986;color:#000\
000;fill:#ffffff\
;fill-opacity:1;\
fill-rule:evenod\
d;stroke:none;st\
roke-width:5;mar\
ker:none;visibil\
ity:visible;disp\
lay:inline;overf\
low:visible;enab\
le-background:ac\
cumulate;filter:\
url(#filter3805)\
\x22 id=\x22path3608\x22 \
sodipodi:cx=\x2211.\
181818\x22 sodipodi\
:cy=\x2220.181818\x22 \
sodipodi:rx=\x2210.\
454545\x22 sodipodi\
:ry=\x227.2727275\x22 \
d=\x22m 21.636363,2\
0.181818 a 10.45\
4545,7.2727275 0\
 1 1 -20.9090900\
1,0 10.454545,7.\
2727275 0 1 1 20\
.90909001,0 z\x22 t\
ransform=\x22matrix\
(0.85867478,-0.7\
467709,0.7057652\
6,0.90856459,-4.\
8393203,997.4188\
)\x22/>\x0d\x0a    <path \
sodipodi:type=\x22a\
rc\x22 style=\x22color\
:#000000;fill:no\
ne;stroke:#2e343\
6;stroke-width:1\
.805642720000000\
03;marker:none;v\
isibility:visibl\
e;display:inline\
;overflow:visibl\
e;enable-backgro\
und:accumulate\x22 \
id=\x22path2826-5\x22 \
sodipodi:cx=\x2231.\
636364\x22 sodipodi\
:cy=\x2230.363636\x22 \
sodipodi:rx=\x2226.\
181818\x22 sodipodi\
:ry=\x2226.181818\x22 \
d=\x22m 57.818182,3\
0.363636 a 26.18\
1818,26.181818 0\
 1 1 -52.363636,\
0 26.181818,26.1\
81818 0 1 1 52.3\
63636,0 z\x22 trans\
form=\x22matrix(1.1\
076389,0,0,1.107\
6389,-3.0416672,\
986.73024)\x22/>\x0d\x0a \
 </g>\x0d\x0a</svg>\x0d\x0a\
"

qt_resource_name = b"\
\x00\x05\
\x00o\xa6S\
\x00i\
\x00c\x00o\x00n\x00s\
\x00\x12\
\x0eO\xfeG\
\x00p\
\x00r\x00e\x00v\x00i\x00e\x00w\x00-\x00v\x00e\x00c\x00t\x00o\x00r\x00.\x00s\x00v\
\x00g\
\x00\x14\
\x02\x9c\x05\x07\
\x00p\
\x00r\x00e\x00v\x00i\x00e\x00w\x00-\x00r\x00e\x00n\x00d\x00e\x00r\x00e\x00d\x00.\
\x00s\x00v\x00g\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x02\x00\x00\x00\x02\
\x00\x00\x00:\x00\x00\x00\x00\x00\x01\x00\x00\x17a\
\x00\x00\x00\x10\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\
"

def qInitResources():
    QtCore.qRegisterResourceData(0x01, qt_resource_struct, qt_resource_name, qt_resource_data)

def qCleanupResources():
    QtCore.qUnregisterResourceData(0x01, qt_resource_struct, qt_resource_name, qt_resource_data)

qInitResources()
