"""Functions and classes that define custom scripted objects.

These classes define a custom object which is based on one of the core
objects defined in C++. The custom object inherits some basic properties,
and new properties are added.

Most Draft objects are based on Part::Part2DObject.
"""
