"""Classes that define the viewproviders of custom scripted objects.

These classes define viewproviders for the custom objects
defined in `draftobjects`.
The viewproviders can be used only when the graphical interface
is available; in console mode the viewproviders are not available.
"""
