"""Utility functions that do not require the graphical user interface.

These functions are used throughout the Draft Workbench.
They can be called from any module, whether it uses the graphical
user interface or not.
"""
