"""Commands that require the graphical user interface to work.

These GUI commands are called by buttons, menus, contextual menus,
toolbars, or other ways that require graphical widgets.
They are normally loaded in the workbench's `InitGui.py`.
"""
